//
//  Element.h
//  oneBuddy
//

#ifndef oneBuddy_Element_h
#define oneBuddy_Element_h
#include "PlaneVector.h"
#include "Utilities.h"
using namespace std;


class Element {
public:
    complex<double> multiplier;
    complex<double> amplitude;
    PlaneVector neighbourVector;
    
    bool isNambu;
    PlaneVector waveVector;
    double phase;
    bool isSpin;
    int band;
    int toband;
    
    Element();
    Element(const Element &);
    Element(const complex<double>&, const PlaneVector&, const bool&, const int&, const int&);
    Element(const complex<double>&, const PlaneVector&, const bool&, const PlaneVector&, const double&, const bool&, const int&, const int&);
    
    void updateAmplitude(const double&);
    void updatePhase(const double&);
    complex<double> value(const PlaneVector&, const int& dx=0, const int& dy=0) const;
    
    bool isNormal() const;
    void forcePeriodXY(int Lx, int Ly);
    void reducePeriodFrac();
    
    complex<double> valueRR(const PlaneVector&, const PlaneVector&, const int&, const int&) const;

};


Element::Element(){
    multiplier = 0;
    amplitude = 0;
    neighbourVector = PlaneVector(0,1,0,1);
    isNambu = false;
    waveVector = PlaneVector(0,1,0,1);
    phase = 0;
    isSpin = false;
    band = 0;
    toband = 0;
}

Element::Element(Element const & t){
    multiplier = t.multiplier;
    amplitude = t.amplitude;
    neighbourVector = t.neighbourVector;
    isNambu = t.isNambu;
    waveVector = t.waveVector;
    phase = t.phase;
    isSpin = t.isSpin;
    band = t.band;
    toband = t.toband;
}

Element::Element(const complex<double> &m, const PlaneVector &a,
                 const bool &nambu=false,
                 const int &bd=0, const int &tbd=0){
    multiplier = m;
    amplitude = m;
    neighbourVector = a;
    isNambu = nambu;
    waveVector = PlaneVector(0,1,0,1);
    phase = 0;
    isSpin = false;
    band = bd;
    toband = tbd;
}

Element::Element(const complex<double> &m, const PlaneVector &a,
                 const bool &nambu, const PlaneVector &Q, const double &phi=0,
                 const bool &spin=false,
                 const int &bd=0, const int &tbd=0){
    multiplier = m;
    amplitude = m;
    neighbourVector = a;
    isNambu = nambu;
    waveVector = Q;
    phase = phi;
    isSpin = spin;
    band = bd;
    toband = tbd;
}

void Element::updateAmplitude(const double &factor) {
    amplitude = multiplier*factor;
}

void Element::updatePhase(const double &factor) {
    phase = factor;
}


void Element::forcePeriodXY(int Lx, int Ly) {
    waveVector.forceDenomXY(Lx,Ly);
}

void Element::reducePeriodFrac() {
    waveVector.reduce();
}

bool Element::isNormal() const{
    bool isNotDW = (waveVector.xNum==0 and waveVector.yNum==0);
    bool isNotNambu = (not isNambu);
    bool isNotSpin = (not isSpin);
    bool isNotInterBand = (band == toband);
    
    return (isNotDW and isNotNambu and isNotSpin and isNotInterBand);
}



complex<double> Element::value(const PlaneVector &k, const int& dx, const int& dy) const{
    complex<double> factor = 1.0;
    complex<double> I(0.0,1.0);
    
    for(int i=0; i<dx; i++) factor *= I*((double)neighbourVector.x);
    for(int i=0; i<dy; i++) factor *= I*((double)neighbourVector.y);
    
    if (isNambu) {
        complex<double> exp = factor*amplitude*polar(1.0, -1.0*(k+M_PI*waveVector)*neighbourVector);
        return (exp+conj(exp))*polar(1.0, 2*M_PI*phase);
    } else {
        return factor*amplitude*(polar(1.0, -1.0*(k+M_PI*waveVector)*neighbourVector+2*M_PI*phase));
    }
}







complex<double> Element::valueRR(const PlaneVector& k, const PlaneVector& R, const int& Lx, const int& Ly) const {
    
    complex<double> valueRR = amplitude;
    int axtilde=0;
    int aytilde=0;
    
    axtilde = floorDiv( R.xNum + neighbourVector.xNum , Lx );
    aytilde = floorDiv( R.yNum + neighbourVector.yNum , Ly );
    valueRR *= polar(1.0, 2*M_PI*waveVector * (R + (+0.5)*neighbourVector));
    
    complex<double> phase_x = polar(1.0, -k.x * Lx * axtilde);
    complex<double> phase_y = polar(1.0, -k.y * Ly * aytilde);
    valueRR *= phase_x * phase_y;
    valueRR *= polar(1.0, 2*M_PI*phase);
    
    return valueRR;
}

















#endif
