//
//  HmMatrix.h (Hermitian Matrix)
//  OneBuddy
//

#ifndef __OneBuddy__Matrix__
#define __OneBuddy__Matrix__

#include <stdio.h>
#include <complex>
#include <exception>
#include <iostream>
#include <set>


#ifndef SMALL_VALUE
#define SMALL_VALUE  double(10e-14)
#endif

using namespace std;

////////// UTILITY for COMPLEX
bool cplxCompare(const complex<double> & z1, const complex<double> & z2, double tolerance = SMALL_VALUE) {
    return ( abs(z1.real()-z2.real()) < tolerance and abs(z1.imag() - z2.imag()) < tolerance );
}

////////// CALL of LAPACK functions (matrix inversion, diagonalization and multiplication)
extern "C" {
    void zgetrf_(int*, int* , complex<double>*, int*, int*, int*);
    void zgetri_(int*, complex<double>*, int*, int*, complex<double>*, int*, int*);
    void zheev_( char*, char*, int*, complex<double>*, int*, double*,complex<double>*, int*, double*, int*);
    void zgemm_( char*, char*, int*, int*, int*, complex<double>*, complex<double>*, int*, complex<double>*, int*, complex<double>*, complex<double>*, int*);
}

////////// LAPACK WRAPPERS
////////// arrays must be in col_major convention : A_{i,j} = A[i+N*j] (because Lapack is in FORTRAN)
void LPKinverse(complex<double>* A, int N) {
    int *IPIV = new int[N+1];
    int LWORK = N*N;
    complex<double> *WORK = new complex<double>[LWORK];
    int INFO;
    zgetrf_(&N,&N,A,&N,IPIV,&INFO);
    zgetri_(&N,A,&N,IPIV,WORK,&LWORK,&INFO);
//    if(INFO > 0) {printf( "LAPACK inversion signals SINGULAR MATRIX.\n" ); exit( 1 );}
//    if(INFO < 0) {printf( "LAPACK signals argument %d is wrong.\n", INFO ); exit( 1 );}
    delete[] IPIV;
    delete[] WORK;
}

void LPKproduct(complex<double>* A, complex<double>* B, complex<double>* Result, int N) {
    complex<double> one(1.0,0.0);
    complex<double> zero(0.0,0.0);
    char char1='N';
    zgemm_( &char1, &char1, &N, &N, &N, &one, A, &N, B, &N, &zero, Result, &N);
}







///////////////////
//// FRIENDS DECLARATION
template <class TYPE> class HmMatrix;
template <class TYPE> void diagonalize(const HmMatrix<TYPE>&, TYPE*, double*);
template <class TYPE> void diagonalize(const HmMatrix<TYPE>&, HmMatrix<TYPE>&, vector<double>&);
template <class TYPE> void diagonalize(const HmMatrix<TYPE>&, HmMatrix<TYPE>&, HmMatrix<TYPE>&);



///////////////////
//// CLASS DECLARATION

template<class TYPE>
class HmMatrix {
    TYPE*     CONTENT;
    size_t    DIM;
    size_t    CAP;
    
    void testRange(int, int) const;
    TYPE& element(int,int);
    const TYPE& element(int,int) const;
    
public:
    HmMatrix();
    ~HmMatrix();
    HmMatrix(size_t);
    HmMatrix(const HmMatrix&);
    void operator=(const HmMatrix&);
    
    const size_t&  dim() const;
    void           assertSameSize(const HmMatrix&) const;
    void           resize(const size_t&);
    void           setToZeros();
    
    TYPE&          operator()(int,int);
    const TYPE&    operator()(int,int)           const;
    void           print()                       const;
    void           printReal()                       const;
    void           printImag()                       const;
    
    TYPE           trace() const;
    TYPE           partialTrace(int,int startIndex=0) const;
    void           squareBloc(HmMatrix&,unsigned int,unsigned int startIndex=0) const;
    HmMatrix&      selfSquareBloc(unsigned int,unsigned int startIndex=0) const;
    void           squareBloc(HmMatrix<TYPE>&, set<int>& ) const;
    
    HmMatrix&      selfImag();
    HmMatrix&      selfConj();
    const HmMatrix conj() const;
    HmMatrix&      selfTranspose();
    const HmMatrix transpose() const;
    HmMatrix&      selfAdjoint();
    const HmMatrix adjoint() const;
    HmMatrix&      selfInverse();
    const HmMatrix inverse() const;
    
    bool           operator==(const HmMatrix&) const;
    bool           operator!=(const HmMatrix&) const;
    
    HmMatrix&      operator*= (const HmMatrix&);
    const HmMatrix operator*  (const HmMatrix&) const;
    HmMatrix&      operator+= (const HmMatrix&);
    const HmMatrix operator+  (const HmMatrix&) const;
    HmMatrix&      operator-= (const HmMatrix&);
    const HmMatrix operator-  (const HmMatrix&) const;
    
    HmMatrix&      operator*=(const TYPE&);
    HmMatrix<TYPE> timesDiag(const vector<double>&) const;
    HmMatrix       timesDiag(const HmMatrix&) const;
    HmMatrix       diagTimes(const HmMatrix&) const;
    
    
    friend void diagonalize<TYPE>(const HmMatrix&, TYPE*, double*);
    friend void diagonalize<TYPE>(const HmMatrix&, HmMatrix&, vector<double>&);
    friend void diagonalize<TYPE>(const HmMatrix&, HmMatrix&, HmMatrix&);
    
    
};

//// Specialized declarations
template<>   void HmMatrix<complex<double>>::print() const ;
template<>   HmMatrix<complex<double>>& HmMatrix<complex<double>>::operator*=(const HmMatrix& rhs);
template<>   const HmMatrix<complex<double>> HmMatrix<complex<double>>::operator*(const HmMatrix& rhs) const;
template<>   bool HmMatrix<complex<double>>::operator==(const HmMatrix& rhs) const;
// UNTESTED
template<>   HmMatrix<complex<double>>& HmMatrix<complex<double>>::selfImag();

///////////////////
//// PRIVATE

template<class TYPE>
void HmMatrix<TYPE>::testRange(int i, int j) const {
    bool inRange = (i >= 0 and i < (int)DIM and j >= 0 and j < (int)DIM);
    if (not inRange)
        throw out_of_range("element does not exist");
}

template<class TYPE>
void HmMatrix<TYPE>::resize(const size_t& dim) {
    if (CAP < dim*dim) {
        if (dim*dim > 2*CAP) CAP = dim*dim;
        else CAP = 2*CAP;
        TYPE* tmpContent = new TYPE[CAP];
        for (size_t i=0;i<DIM*DIM;i++) tmpContent[i] = TYPE(CONTENT[i]);
        delete[] CONTENT;
        CONTENT=tmpContent;
    }
    DIM = dim;
}

template<class TYPE>
TYPE& HmMatrix<TYPE>::element(int i,int j) {
    return CONTENT[i+DIM*j]; // col_major convention, for LAPACK
}

template<class TYPE>
const TYPE& HmMatrix<TYPE>::element(int i,int j) const {
    return CONTENT[i+DIM*j];
}


///////////////////
//// PUBLIC

template<class TYPE>
HmMatrix<TYPE>::HmMatrix() {
    DIM=CAP=0;
    CONTENT = nullptr;
}

template<class TYPE>
HmMatrix<TYPE>::HmMatrix(const HmMatrix& rhs) {
    DIM=CAP=0;
    CONTENT = nullptr;
    *this = rhs;
}

template<class TYPE>
HmMatrix<TYPE>::HmMatrix(size_t d){
    DIM = d;
    CAP = d*d;
    CONTENT = new complex<double>[CAP]();
}

template<class TYPE>
HmMatrix<TYPE>::~HmMatrix() {
    delete[] CONTENT;
    CONTENT = nullptr;
    DIM=CAP=0;
}

template<class TYPE>
void HmMatrix<TYPE>::operator=(const HmMatrix& source){
    resize(source.DIM);
    for (size_t i=0; i<DIM*DIM; i++){
        CONTENT[i] = TYPE(source.CONTENT[i]);
    }
}




template<class TYPE>
const size_t& HmMatrix<TYPE>::dim() const{
    return DIM;
}

template<class TYPE>
TYPE& HmMatrix<TYPE>::operator()(int i,int j){
    testRange(i,j);
    return element(i,j);
}

template<class TYPE>
const TYPE& HmMatrix<TYPE>::operator()(int i,int j) const{
    testRange(i,j);
    return element(i,j);
}

template<class TYPE>
void HmMatrix<TYPE>::assertSameSize(const HmMatrix& rhs) const {
    if (DIM != rhs.dim()) throw length_error("matrices must be same size");
}

template<class TYPE>
void HmMatrix<TYPE>::print() const {
    for( size_t  i = 0; i < DIM; i++ ) {
    for( size_t  j = 0; j < DIM; j++ ) {
        if ( abs(element(i,j)) < 0.0001 ) printf( "   .    ");
        else printf( " % 4.6f", element(i,j) );
    } printf( "\n" );
    } printf( "\n\n" );
    fflush(stdout);
}

template<>
void HmMatrix<complex<double>>::print() const {
    for( size_t  i = 0; i < DIM; i++ ) {
    for( size_t  j = 0; j < DIM; j++ ) {
        if ( abs(real(element(i,j))) < 0.0001 and abs(imag(element(i,j))) < 0.0001 ) printf( "       .        ");
        else printf( " (% 5.3f,% 5.3f)", real(element(i,j)), imag(element(i,j)) );
        //printf("%f + i%f ", real(element(i,j)), imag(element(i,j)));
    } printf( "\n" );
    } printf( "\n\n" );
    fflush(stdout);
}

template<>
void HmMatrix<bool>::print() const {
    for( size_t  i = 0; i < DIM; i++ ) {
    for( size_t  j = 0; j < DIM; j++ ) {
        if ( element(i,j) ) printf( "1 ");
        else printf( ". ");
    } printf( "\n" );
    } printf( "\n\n" );
    fflush(stdout);
}

template<>
void HmMatrix<complex<double>>::printReal() const {
    for( size_t  i = 0; i < DIM; i++ ) {
    for( size_t  j = 0; j < DIM; j++ ) {
        if ( abs(real(element(i,j))) < 0.0001 and abs(imag(element(i,j))) < 0.0001 ) printf( "  .   ");
        else printf( "% 4.2f ", real(element(i,j)) );
    } printf( "\n" );
    } printf( "\n\n" );
    fflush(stdout);
}

template<>
void HmMatrix<complex<double>>::printImag() const {
    for( size_t  i = 0; i < DIM; i++ ) {
    for( size_t  j = 0; j < DIM; j++ ) {
        if ( abs(real(element(i,j))) < 0.0001 and abs(imag(element(i,j))) < 0.0001 ) printf( "  .   ");
        else printf( "% 4.2f ", imag(element(i,j)) );
    } printf( "\n" );
    } printf( "\n\n" );
    fflush(stdout);
}



template<class TYPE>
TYPE HmMatrix<TYPE>::trace() const {
    TYPE result=0;
    for (size_t i=0; i<DIM; i++)
        result += element(i,i);
    return result;
}

template<class TYPE>
HmMatrix<TYPE>& HmMatrix<TYPE>::selfConj() {
    for (size_t i=0; i<DIM*DIM; i++)
        CONTENT[i] = std::conj(CONTENT[i]);
    return *this;
}

template<class TYPE>
const HmMatrix<TYPE> HmMatrix<TYPE>::conj() const{
    return HmMatrix<TYPE>(*this).selfConj();
}

template<class TYPE>
HmMatrix<TYPE>& HmMatrix<TYPE>::selfTranspose() {
    if (DIM > 1){
        TYPE tmp;
        for (size_t i=0; i<DIM-1; i++) {
        for (size_t j=i+1; j<DIM; j++){
            tmp = element(i,j);
            element(i,j) = element(j,i);
            element(j,i) = tmp;
        }}
    }
    return *this;
}

template<class TYPE>
const HmMatrix<TYPE> HmMatrix<TYPE>::transpose() const{
    return HmMatrix<TYPE>(*this).selfTranspose();
}

template<class TYPE>
HmMatrix<TYPE>& HmMatrix<TYPE>::selfAdjoint() {
    selfConj();
    selfTranspose();
    return *this;
}

template<class TYPE>
const HmMatrix<TYPE> HmMatrix<TYPE>::adjoint() const{
    return HmMatrix<TYPE>(*this).selfAdjoint();
}

template<class TYPE>
HmMatrix<TYPE>& HmMatrix<TYPE>::selfInverse(){
    cerr << "ERROR : HmMatrix inversion only implemented complex<double>" << endl; exit(8);
    return *this;
}

template<>
HmMatrix<complex<double>>& HmMatrix<complex<double>>::selfInverse(){
    if (DIM > 0)
        LPKinverse(CONTENT, int(DIM));
    return *this;
}

template<class TYPE>
const HmMatrix<TYPE> HmMatrix<TYPE>::inverse() const{
    return HmMatrix<TYPE>(*this).selfInverse();
}




template<class TYPE>
HmMatrix<TYPE>& HmMatrix<TYPE>::operator*=(const HmMatrix& rhs){
    cerr << "ERROR : HmMatrix multiplication only implemented complex<double>" << endl; exit(8);
    return *this;
}

template<>
HmMatrix<complex<double>>& HmMatrix<complex<double>>::operator*=(const HmMatrix& rhs){
    assertSameSize(rhs);
    complex<double>* resultContent = new complex<double>[CAP];
    LPKproduct(CONTENT, rhs.CONTENT, resultContent, int(DIM));
    delete[] CONTENT;
    CONTENT = resultContent;
    return *this;
}

template<class TYPE>
const HmMatrix<TYPE> HmMatrix<TYPE>::operator*(const HmMatrix& rhs) const{
    return HmMatrix<TYPE>(*this) *= rhs;
}

template<>  // one less allocation than using *=
const HmMatrix<complex<double>> HmMatrix<complex<double>>::operator*(const HmMatrix& rhs) const{
    HmMatrix<complex<double>> resultMatrix(DIM);
    assertSameSize(rhs);
    LPKproduct(CONTENT, rhs.CONTENT, resultMatrix.CONTENT, int(DIM));
    return resultMatrix;
}

template<class TYPE>
HmMatrix<TYPE>& HmMatrix<TYPE>::operator+=(const HmMatrix& rhs){
    assertSameSize(rhs);
    for (size_t i=0; i<DIM*DIM; i++)
        CONTENT[i] += rhs.CONTENT[i];
    return *this;
}

template<class TYPE>
const HmMatrix<TYPE> HmMatrix<TYPE>::operator+(const HmMatrix& rhs) const{
    return HmMatrix<TYPE>(*this) += rhs;
}

template<class TYPE>
HmMatrix<TYPE>& HmMatrix<TYPE>::operator-=(const HmMatrix& rhs){
    assertSameSize(rhs);
    for (size_t i=0; i<DIM*DIM; i++)
        CONTENT[i] -= rhs.CONTENT[i];
    return *this;
}

template<class TYPE>
const HmMatrix<TYPE> HmMatrix<TYPE>::operator-(const HmMatrix& rhs) const{
    return HmMatrix<TYPE>(*this) -= rhs;
}









template<class TYPE>
bool HmMatrix<TYPE>::operator==(const HmMatrix& rhs) const {
    cerr << "ERROR : HmMatrix boolean comparisons only implemented complex<double>" << endl; exit(8);
    return false;
}

template<>
bool HmMatrix<complex<double>>::operator==(const HmMatrix& rhs) const {
    bool sameSoFar = DIM==rhs.DIM;
    int i=0;
    while (sameSoFar and i < (int)(DIM*DIM)) {
        sameSoFar = cplxCompare(CONTENT[i],rhs.CONTENT[i]);
        i++;
    }
    return sameSoFar;
}
    
template<class TYPE>
bool HmMatrix<TYPE>::operator!=(const HmMatrix& rhs) const {
    return not operator==(rhs);
}




template<class TYPE>
HmMatrix<TYPE>& HmMatrix<TYPE>::operator*=(const TYPE& z){
    for (size_t i=0; i<DIM*DIM; i++) CONTENT[i]*=z;
    return *this;
}




template<class TYPE>
HmMatrix<TYPE> HmMatrix<TYPE>::timesDiag(const HmMatrix& rhs) const {
    HmMatrix<TYPE> output(DIM);
    assertSameSize(rhs);
    for (size_t i=0; i<DIM; ++i) {
        for (size_t j=0; j<DIM; ++j) {
            output.element(i,j) += element(i,j)*rhs.element(j,j);
        }
    }
    return output;
}

template<class TYPE>
HmMatrix<TYPE> HmMatrix<TYPE>::diagTimes(const HmMatrix& rhs) const {
    HmMatrix<TYPE> output(DIM);
    assertSameSize(rhs);
    for (size_t i=0; i<DIM; ++i) {
        for (size_t j=0; j<DIM; ++j) {
            output.element(i,j) += element(i,i)*rhs.element(i,j);
        }
    }
    return output;
}


template<class TYPE>
HmMatrix<TYPE> HmMatrix<TYPE>::timesDiag(const vector<double>& rhs) const {
    HmMatrix<TYPE> output(DIM);
    if (DIM != rhs.size()) throw length_error("matrices must be same size as the vector");
    for (size_t i=0; i<DIM; ++i) {
        for (size_t j=0; j<DIM; ++j) {
            output.element(i,j) += element(i,j)*rhs[j];
        }
    }
    return output;
}



template<class TYPE>
void HmMatrix<TYPE>::setToZeros() {
    for (size_t i=0; i<DIM*DIM; ++i) {
        CONTENT[i]=(TYPE) 0.0;
    }
}


template<class TYPE>
HmMatrix<TYPE>& HmMatrix<TYPE>::selfImag() {
    return *this;
}

template<>
HmMatrix<complex<double>>& HmMatrix<complex<double>>::selfImag() {
    for (size_t i=0; i<(DIM*DIM); ++i){
        complex<double> tmp = imag(CONTENT[i]);
        CONTENT[i] = tmp;
    }
    return *this;
}

template<class TYPE>
TYPE HmMatrix<TYPE>::partialTrace(int lowerDim, int startIndex) const { /// default startIndex = 0 (see declaration)
    TYPE result=0;
    int endIndex = startIndex+lowerDim ;
    testRange(endIndex-1,endIndex-1);
    for (int i=startIndex; i<endIndex; i++)
        result += element(i,i);
    return result;
}




///////////////////
//// FRIENDS AND INLINE

template<class TYPE>
inline const HmMatrix<TYPE> operator*(const HmMatrix<TYPE>& M, const TYPE& z){
    return HmMatrix<TYPE>(M)*=z;
}

template<class TYPE>
inline HmMatrix<TYPE> operator*(const TYPE& z, const HmMatrix<TYPE>& M){
    return M*z;
}

inline HmMatrix<complex<double>> unit(const size_t& dim) {
    HmMatrix<complex<double>> output(dim);
    for (size_t i=0; i< dim; i++) output(i,i)=1.0;
    return output;
}

inline const HmMatrix<complex<double>> operator*(const HmMatrix<complex<double>>& M, const double& z){
    return HmMatrix<complex<double>>(M)*=complex<double>(z);
}

inline HmMatrix<complex<double>> operator*(const double& z, const HmMatrix<complex<double>>& M){
    return M*z;
}






template<class TYPE>
void HmMatrix<TYPE>::squareBloc(HmMatrix<TYPE>& outPutMatrix, set<int>& indices) const {
    if (outPutMatrix.DIM != indices.size()){
        outPutMatrix.resize(indices.size());
    }
    int i0=0;
    int j0=0;
    for(set<int>::iterator i1=indices.begin(); i1!=indices.end();++i1)
    {   
        j0=0;
        for(set<int>::iterator j1=indices.begin(); j1!=indices.end();++j1)
        {
            size_t i = *i1;
            size_t j = *j1;
            //printf("i0:%d j0:%d i:%d j:%d DIM:%d \n", i0, j0, i, j, this->DIM); fflush(stdout);
            if ((i > this->DIM) || (i<0) || (j > this->DIM) || (j<0))  throw runtime_error("The dimension of the set exceed the dimension of the matrix.");
            outPutMatrix.CONTENT[i0+indices.size()*j0] = this->CONTENT[i+(this->DIM)*j];
            j0++;
        }
        i0++;
    }
}


template<class TYPE>
void HmMatrix<TYPE>::squareBloc(HmMatrix<TYPE>& outPutMatrix,unsigned int size,unsigned int startIndex) const {
    if (outPutMatrix.DIM != size){
        outPutMatrix.resize(size);
    }
    if ((startIndex+size > this->DIM) || (size==0)) {
        cerr << "The dimension of the subset exceed the dimension of the matrix." << endl; exit(8);
    }
    for (unsigned int j=0; j<size; ++j){
        for (unsigned int i=0; i<size; ++i){
            outPutMatrix.CONTENT[i+size*j] = this->CONTENT[(i+startIndex)+(this->DIM)*(j+startIndex)];
        }
    }
}

template<class TYPE>
HmMatrix<TYPE>& HmMatrix<TYPE>::selfSquareBloc(unsigned int size,unsigned int startIndex) const{
    if ((startIndex+size > this->DIM) || (size==0)) {
        cerr << "The dimension of the subset exceed the dimension of the matrix." << endl; exit(8);
    }
    for (unsigned int j=0; j<size; ++j){  //Caution, the order is important here. Do j before i.
        for (unsigned int i=0; i<size; ++i){
            this->CONTENT[i+size*j] = this->CONTENT[(i+startIndex)+(this->DIM)*(j+startIndex)];
        }
    }
    this->resize(size);
}


void LPKeigenSolve(complex<double>* matrixInAndOutEigenVec, double* outEigenVal, int DIM) {
    int INFO, LWORK = DIM*DIM;
    complex<double> *WORK = new complex<double>[LWORK];
    double RWORK[3*DIM-2];
    char V='V', L='L';
    
    zheev_( &V, &L, &DIM, matrixInAndOutEigenVec, &DIM, outEigenVal, WORK, &LWORK, RWORK, &INFO );
    if( !(INFO == 0) ) {
        printf( "The algorithm failed to compute eigenvalues.\n" );
        exit( 1 );
    }
    
    delete[] WORK;
}









////////////////////////////////////////////////////////////////////////////////


template <class TYPE>
void diagonalize(const HmMatrix<TYPE>& inputMatrix, TYPE* outEigenVectors, double* outEigenVals) {
    cerr << "ERROR : HmMatrix diagonalize only implemented for complex<double>" << endl; exit(8);
}

template <class TYPE>
void diagonalize(const HmMatrix<TYPE>& inputMatrix, HmMatrix<TYPE>& outEigenVectors, vector<double>& outEigenVals) {
    cerr << "ERROR : HmMatrix diagonalize only implemented for complex<double>" << endl; exit(8);
}

// ONLY USEFUL FOR TESTS MAY REMOVE
template <class TYPE>
void diagonalize(const HmMatrix<TYPE>& inputMatrix, HmMatrix<TYPE>& outEigenVectors, HmMatrix<TYPE>& outEigenVals) {
    cerr << "ERROR : HmMatrix diagonalize only implemented for complex<double>" << endl; exit(8);
}


// The order of the declaration is important here forward declarations would be more intelligent
template <>
void diagonalize (const HmMatrix<complex<double>>& inputMatrix, complex<double>* outEigenVectors, double* outEigenVals) {
    for (unsigned int i=0; i<inputMatrix.DIM*inputMatrix.DIM; ++i){
        outEigenVectors[i] = inputMatrix.CONTENT[i];
    }
    LPKeigenSolve(outEigenVectors, outEigenVals, (int)inputMatrix.DIM);
}

template <>
void diagonalize (const HmMatrix<complex<double>>& inputMatrix, HmMatrix<complex<double>>& outEigenVectors, vector<double>& outEigenVals) {
    inputMatrix.assertSameSize(outEigenVectors);
    if (inputMatrix.dim() != outEigenVals.size()) throw length_error("the matrix must be same size as the eigenvalue vector");
    diagonalize(inputMatrix, outEigenVectors.CONTENT, &outEigenVals[0]);
}

// ONLY USEFUL FOR TESTS MAY REMOVE
template <>
void diagonalize (const HmMatrix<complex<double>>& inputMatrix, HmMatrix<complex<double>>& outEigenVectors, HmMatrix<complex<double>>& outEigenVals) {
    vector<double> tmpEigenVal(inputMatrix.DIM);
    diagonalize(inputMatrix, outEigenVectors, tmpEigenVal);
    
    //// Should be a method of matrix diag()
    outEigenVals.setToZeros();
    for (unsigned int i=0; i<outEigenVals.DIM; ++i){
        outEigenVals(i,i) = tmpEigenVal[i];
    }
}


////////////////////////////////////////////////////////////////////////////////



#endif /* defined(__OneBuddy__Matrix__) */
