//
//  testMatrix.cpp
//  OneBody
//
//  Created by Simon Verret on 2016-07-29.
//  Copyright (c) 2016 Simon Verret. All rights reserved.
//

#include <stdio.h>
#include "catch.hpp"
#include "../src/HmMatrix.h"
complex<double> cplx0(0,0);
complex<double> I(0,1);

bool cplxApprox(const complex<double> & z1, const complex<double> & z2) {
    return ( z1.real()==Approx(z2.real()) and z1.imag()==Approx(z2.imag()) );
}

bool matrixApprox(const HmMatrix<complex<double>>& M1, const HmMatrix<complex<double>>& M2) {
    bool sameSoFar=true;
    for (int i=0; i<M1.dim(); i++) {
    for (int j=0; j<M1.dim(); j++) {
        if (not cplxApprox(M1(i,j),M2(i,j))) sameSoFar=false;
    }}
    return sameSoFar;
}

TEST_CASE( "Initialize Matrix empty constructor" ) {
    HmMatrix<complex<double>> M;
    
    SECTION( "Matrix size is zero" ) {
        CHECK(M.dim() == 0);
    }
    
    SECTION( "cannot access elements because there is not" ){
        CHECK_THROWS_AS( M(0,0) , out_of_range );
        CHECK_THROWS_AS( M(1,0) , out_of_range );
        CHECK_THROWS_AS( M(0,1) , out_of_range );
        CHECK_THROWS_AS( M(23,17) , out_of_range );
    }
}

TEST_CASE( "Initialize Matrix of a given size, without content") {
    HmMatrix<complex<double>> M(23);
    
    SECTION( "Matrix has right size" ) {
        CHECK(M.dim() == 23);
    }
    
    SECTION( "The Matrix is filled with zeros" ) {
        bool zerosSoFar = true;
        for (int i=0; i<23; i++){
            for (int j=0; j<23; j++){
                zerosSoFar = zerosSoFar and M(i,j)==complex<double>(0);
            }
        }
        CHECK(zerosSoFar);
    }
    
    SECTION( "Acessing elements outside range throws exception" ) {
        CHECK_THROWS_AS( M(24,5) , out_of_range );
        CHECK_THROWS_AS( M(6,38) , out_of_range );
        CHECK_THROWS_AS( M(-2,18) , out_of_range );
        CHECK_THROWS_AS( M(22,-30) , out_of_range );
    }
}


TEST_CASE( "one Matrix with specific elements and its copies" ) {
    HmMatrix<complex<double>> M(3);
    for (int i=0;i<3;i++) {
    for (int j=0;j<3;j++) {
        M(i,j) = 3*i+j+4.0;
    }}
    // 4  5  6
    // 7  8  9
    // 10 11 12
    
    HmMatrix<complex<double>> M2=M;
    HmMatrix<complex<double>> M3(M);
    const HmMatrix<complex<double>> M4=M;
    const HmMatrix<complex<double>> M5(M);
    
    SECTION( "Matrix still has the right size" ){
        CHECK(M.dim() == 3);
    }
    
    SECTION( "Matrix has the right content") {
        for (int i=0;i<3;i++) {
        for (int j=0;j<3;j++) {
            CHECK(M(i,j) == 3*i+j+4.0);
        }}
    }
    
    SECTION( "Copy of Matrix has the right size" ) {
        CHECK(M2.dim() == 3);
        CHECK(M3.dim() == 3);
    }
    
    SECTION( "Copy of Matrix has the right content") {
        for (int i=0;i<3;i++) {
        for (int j=0;j<3;j++) {
            CHECK(M2(i,j) == 3*i+j+4.0);
            CHECK(M3(i,j) == 3*i+j+4.0);
        }}
    }
    
    SECTION( "const Copy of Matrix is accessed and has the right content") {
     // because M(i,j) returns a lreference that may not be used as const
     // if badly implemented
        for (int i=0;i<3;i++) {
        for (int j=0;j<3;j++) {
            CHECK(M4(i,j) == 3*i+j+4.0);
            CHECK(M5(i,j) == 3*i+j+4.0);
        }}
    }
}


TEST_CASE( "Valid mathematical operations on Matrices (complex)" ) {
    // ARBITRARY MATRICES
    HmMatrix<complex<double>> A(3);
    A(0,0) =  7.0; A(0,1) =  2.0; A(0,2) =  1.0;
    A(1,0) =  0.0; A(1,1) =  3.0; A(1,2) = -1.0;
    A(2,0) = -1.0; A(2,1) = -2.0; A(2,2) =  4.0;
    HmMatrix<complex<double>> B(3);
    B(0,0) = -2.0; B(0,1) =  5.0; B(0,2) =  1.0;
    B(1,0) =  7.0; B(1,1) =  0.0; B(1,2) =  9.0;
    B(2,0) =  0.0; B(2,1) =  1.0; B(2,2) = -4.0;
    HmMatrix<complex<double>> C(2);
    C(0,0) =  7.0+I*1.0; C(0,1) =  2.0+I*4.0;
    C(1,0) =  0.0+I*2.0; C(1,1) =  3.0-I*3.0;
    HmMatrix<complex<double>> Acopy(A);
    HmMatrix<complex<double>> Bcopy(B);
    HmMatrix<complex<double>> Ccopy(C);
    
    SECTION( "Boolean comparisons as expected") {
        CHECK_THROWS_AS(A.assertSameSize(C), length_error);
        CHECK(A == Acopy);
        CHECK(not (A == B));
        CHECK(not (A == C));
        CHECK(A != B);
        CHECK(A != C);
        CHECK(not (A != Acopy));
    }
    
    SECTION( "setting to zeros" ) {
        A.setToZeros();
        HmMatrix<complex<double>> Expected(3);
        Expected(0,0) = 0; Expected(0,1) = 0; Expected(0,2) = 0;
        Expected(1,0) = 0; Expected(1,1) = 0; Expected(1,2) = 0;
        Expected(2,0) = 0; Expected(2,1) = 0; Expected(2,2) = 0;
        
        CHECK(matrixApprox(A, Expected));
    }
    
    SECTION( "Transpose as expected" ) {
        HmMatrix<complex<double>> AT = A.transpose();
        Acopy.selfTranspose();
        
        for (int i=0; i<AT.dim(); i++) {
        for (int j=0; j<AT.dim(); j++){
            CHECK(A(i,j) == AT(j,i));
        }}
        
        for (int i=0; i<Acopy.dim(); i++) {
        for (int j=0; j<Acopy.dim(); j++){
            CHECK(A(i,j) == Acopy(j,i));
        }}
    }
    
    SECTION( "Conjuguate as expected" ) {
        HmMatrix<complex<double>> Cstar = C.conj();
        Ccopy.selfConj();
        
        for (int i=0; i<Cstar.dim(); i++) {
            for (int j=0; j<Cstar.dim(); j++){
                CHECK(C(i,j) == conj(Cstar(i,j)));
            }}
        
        for (int i=0; i<Ccopy.dim(); i++) {
            for (int j=0; j<Ccopy.dim(); j++){
                CHECK(C(i,j) == conj(Ccopy(i,j)));
            }}
    }
    
    SECTION( "Adjoint as expected" ) {
        HmMatrix<complex<double>> Cdag = C.adjoint();
        Ccopy.selfAdjoint();
        
        for (int i=0; i<Cdag.dim(); i++) {
            for (int j=0; j<Cdag.dim(); j++){
                CHECK(C(i,j) == conj(Cdag(j,i)));
            }}
        
        for (int i=0; i<Ccopy.dim(); i++) {
            for (int j=0; j<Ccopy.dim(); j++){
                CHECK(C(i,j) == conj(Ccopy(j,i)));
            }}
    }
    
    SECTION( "Imaginary part as expected" ) {
        C.selfImag();
        HmMatrix<complex<double>> Expected(2);
        Expected(0,0) =  1; Expected(0,1) =  4;
        Expected(1,0) =  2; Expected(1,1) = -3;
        CHECK( matrixApprox( C, Expected));
    }
    
    SECTION( "Trace as expected" ) {
        CHECK( A.trace() == 14.0 );
        CHECK( B.trace() == -6.0 );
        CHECK( C.trace() == complex<double>(10,-2));
    }
    
    SECTION( "Partial trace as expected" ) {
        CHECK( A.partialTrace(1) == 7.0 );
        CHECK( A.partialTrace(2) == 10.0 );
        CHECK( A.partialTrace(3) == 14.0 );
        CHECK( B.partialTrace(2) == -2.0 );
        CHECK( C.partialTrace(1) == complex<double>(7,1));
    }
    
    SECTION( "Inverse as expected" ) {
        Acopy.selfInverse();
        HmMatrix<complex<double>> Expected(3);
        Expected(0,0) =  0.1333333; Expected(0,1) = -0.1333333; Expected(0,2) = -0.0666667;
        Expected(1,0) =  0.0133333; Expected(1,1) =  0.3866667; Expected(1,2) =  0.0933333;
        Expected(2,0) =       0.04; Expected(2,1) =       0.16; Expected(2,2) =       0.28;
        CHECK( matrixApprox( A.inverse(), Expected));
        CHECK( matrixApprox( Acopy, Expected));
    }
    
    SECTION( "Multiplication as expected" ) {
        HmMatrix<complex<double>> Expected(3);
        Expected(0,0) =   0; Expected(0,1) = 36; Expected(0,2) =  21;
        Expected(1,0) =  21; Expected(1,1) = -1; Expected(1,2) =  31;
        Expected(2,0) = -12; Expected(2,1) = -1; Expected(2,2) = -35;
        CHECK( matrixApprox( A*B, Expected));
    }
    
    SECTION( "self multiplication as expected" ) {
        HmMatrix<complex<double>> Expected(3);
        Expected(0,0) =   48; Expected(0,1) = 18; Expected(0,2) =  9;
        Expected(1,0) =  1; Expected(1,1) = 11; Expected(1,2) =  -7;
        Expected(2,0) = -11; Expected(2,1) = -16; Expected(2,2) = 17;
        CHECK( matrixApprox( A*A, Expected));
    }
    
    
    SECTION( "multiplication with diagonal only" ) {
        HmMatrix<complex<double>> Expected(3);
        
        Expected(0,0) = -14.; Expected(0,1) =  0.; Expected(0,2) =  -4.;
        Expected(1,0) =   0.; Expected(1,1) =  0.; Expected(1,2) =   4.;
        Expected(2,0) =   2.; Expected(2,1) =  0.; Expected(2,2) = -16.;
        CHECK( matrixApprox( A.timesDiag(B), Expected));
        
        Expected(0,0) = -14; Expected(0,1) = 35; Expected(0,2) =   7;
        Expected(1,0) =  21; Expected(1,1) =  0; Expected(1,2) =  27;
        Expected(2,0) =   0; Expected(2,1) =  4; Expected(2,2) = -16;
        CHECK( matrixApprox( A.diagTimes(B), Expected));
        
        Expected(0,0) = -14; Expected(0,1) = 15; Expected(0,2) =   4;
        Expected(1,0) =  49; Expected(1,1) =  0; Expected(1,2) =  36;
        Expected(2,0) =   0; Expected(2,1) =  3; Expected(2,2) = -16;
        CHECK( matrixApprox( B.timesDiag(A), Expected));
        
        Expected(0,0) = -14; Expected(0,1) =  -4; Expected(0,2) =  -2;
        Expected(1,0) =   0; Expected(1,1) =   0; Expected(1,2) =   0;
        Expected(2,0) =   4; Expected(2,1) =   8; Expected(2,2) = -16;
        CHECK( matrixApprox( B.diagTimes(A), Expected));
        
        Expected(0,0) =  49; Expected(0,1) =  14; Expected(0,2) =  7;
        Expected(1,0) =   0; Expected(1,1) =   9; Expected(1,2) = -3;
        Expected(2,0) =  -4; Expected(2,1) =  -8; Expected(2,2) = 16;
        CHECK( matrixApprox( A.diagTimes(A), Expected));
        
        Expected(0,0) =  49; Expected(0,1) =   6; Expected(0,2) =  4;
        Expected(1,0) =   0; Expected(1,1) =   9; Expected(1,2) = -4;
        Expected(2,0) =  -7; Expected(2,1) =  -6; Expected(2,2) = 16;
        CHECK( matrixApprox( A.timesDiag(A), Expected));
    }
    
    SECTION( "Multiplication with complex as expected" ) {
        complex<double> z(2.0,3.0);
        HmMatrix<complex<double>> Expected(3);
        Expected(0,0) =  7.0*z; Expected(0,1) =  2.0*z; Expected(0,2) =  1.0*z;
        Expected(1,0) =  0.0*z; Expected(1,1) =  3.0*z; Expected(1,2) = -1.0*z;
        Expected(2,0) = -1.0*z; Expected(2,1) = -2.0*z; Expected(2,2) =  4.0*z;
        CHECK( matrixApprox( z*A, Expected));
        CHECK( matrixApprox( A*z, Expected));
    }
    
    SECTION( "Addition as expected" ) {
        Acopy += B;
        HmMatrix<complex<double>> Expected(3);
        Expected(0,0) =  5; Expected(0,1) =  7; Expected(0,2) =  2;
        Expected(1,0) =  7; Expected(1,1) =  3; Expected(1,2) =  8;
        Expected(2,0) = -1; Expected(2,1) = -1; Expected(2,2) =  0;
        CHECK( matrixApprox( A+B, Expected) );
        CHECK( matrixApprox( Acopy, Expected) );
    }
    
    SECTION( "Addition and substraction as expected" ) {
        Acopy -= B;
        HmMatrix<complex<double>> Expected(3);
        Expected(0,0) =  9; Expected(0,1) = -3; Expected(0,2) =   0;
        Expected(1,0) = -7; Expected(1,1) =  3; Expected(1,2) = -10;
        Expected(2,0) = -1; Expected(2,1) = -3; Expected(2,2) =   8;
        CHECK( matrixApprox( A-B, Expected) );
        CHECK( matrixApprox( Acopy, Expected) );
    }
}

TEST_CASE( "Handling empty, size=1 and different size matrices" ) {
    // Tests failures in this section can be pretty Nasty,
    // expect segmentation faults, infinite loops, malloc errors, etc.
    // instead of simple test failures
    
    // ARBITRARY MATRICES
    HmMatrix<complex<double>> A(3);
    A(0,0) =  7.0; A(0,1) =  2.0; A(0,2) =  1.0;
    A(1,0) =  0.0; A(1,1) =  3.0; A(1,2) = -1.0;
    A(2,0) = -1.0; A(2,1) = -2.0; A(2,2) =  4.0;
    HmMatrix<complex<double>> C(2);
    C(0,0) =  7.0+I*1.0; C(0,1) =  2.0+I*4.0;
    C(1,0) =  0.0+I*2.0; C(1,1) =  3.0-I*3.0;
    HmMatrix<complex<double>> Acopy(A);
    HmMatrix<complex<double>> Ccopy(C);
    
    HmMatrix<complex<double>> D(1); D(0,0)=5.0+I*1.0;
    HmMatrix<complex<double>> Dcopy(D);
    
    HmMatrix<complex<double>> E;
    HmMatrix<complex<double>> Ecopy(E);
    
    SECTION( "Boolean with empty matrix") {
        CHECK(D == D);
        CHECK_FALSE(D == E);
        CHECK(D != E);
        CHECK_FALSE(A == E);
        CHECK(D != E);
        CHECK(E == E);
        CHECK_FALSE(A == E);
        CHECK(A != E);
    }
    
    SECTION( "Transpose of empty matrix" ) {
        CHECK( Ecopy == E.transpose());
        CHECK( Ecopy == E.selfTranspose());
    }
    
    SECTION( "Conjuguate of empty matrix" ) {
        CHECK( Ecopy == E.conj());
        CHECK( Ecopy == E.selfConj());
    }
    
    SECTION( "Adjoint of empty matrix" ) {
        CHECK( Ecopy == E.adjoint());
        CHECK( Ecopy == E.selfAdjoint());
    }
    
    SECTION( "Trace of empty matrix" ) {
        CHECK( E.trace() == 0.0 );
    }
    
    SECTION( "Inverse of empty matrix" ) {
        CHECK( Ecopy == E.inverse());
        CHECK( Ecopy == E.selfInverse());
    }
    
    SECTION( "Multiplication as expected" ) {
        CHECK_THROWS_AS( A*C, length_error );
        CHECK_THROWS_AS( A*E, length_error );
    }
    
    SECTION( "Addition and substraction as expected" ) {
        CHECK_THROWS_AS( A+C, length_error );
        CHECK_THROWS_AS( A+E, length_error );
        CHECK_THROWS_AS( A-C, length_error );
        CHECK_THROWS_AS( A-E, length_error );
    }
}


TEST_CASE( "Unitary matrix" ) {
    HmMatrix<complex<double>> unit3 = unit(3);
    HmMatrix<complex<double>> expected(3);
    expected(0,0) = 1 ; expected(0,1) = 0 ; expected(0,2) = 0 ;
    expected(1,0) = 0 ; expected(1,1) = 1 ; expected(1,2) = 0 ;
    expected(2,0) = 0 ; expected(2,1) = 0 ; expected(2,2) = 1 ;
    
    CHECK( matrixApprox(unit3, expected) );
}



//// Friend functions
TEST_CASE( "Eigenvalues and Eigenvectors" ){
    // Private function diagonalize() computes and stocks the eigenvectors and
    // eigenvalues within the HmMatrix object when necessary. As long as the
    // matrix is unchanged, they can be retrieved without recomputing. Modifiers
    // (+=, -=, *=, selfInverse(), etc.) will reset the stored values.
    
    // ARBITRARY MATRIX ONLY WORKS FOR HERMITIAN MATRIX
    HmMatrix<complex<double>> A(3);
    A(0,0) =  1.0;      A(0,1) =  2.0;   A(0,2) =  1.0-I*2.0;
    A(1,0) =  2.0;      A(1,1) =  3.0;   A(1,2) = -3.0+I;
    A(2,0) = 1.0+I*2.0; A(2,1) = -3.0-I; A(2,2) =  12.0;
    
    // CHECKED in NUMPY and MATHEMATICA
    HmMatrix<complex<double>> expctdVal(3);
    expctdVal(0,0) = -1.11559; expctdVal(0,1) =    0.0; expctdVal(0,2) =     0.0;
    expctdVal(1,0) =      0.0; expctdVal(1,1) = 3.85228; expctdVal(1,2) =     0.0;
    expctdVal(2,0) =      0.0; expctdVal(2,1) =    0.0; expctdVal(2,2) = 13.2633;
    
    // WORKED ONCE FOR INTERNAL CONSISTENCY (CANNOT COMPARE WITH NUMPY OR MATHEMATICA : DEFINED UP TO A PHASE)
    HmMatrix<complex<double>> expctdVec(3);
    expctdVec(0,0) = 0.784642 ; expctdVec(0,1) = 0.602013 ; expctdVec(0,2) = -0.148042;
    expctdVec(1,0) = -0.557212-0.089208*I ; expctdVec(1,1) = 0.756612+0.177646*I ; expctdVec(1,2) = 0.123465+0.249587*I ;
    expctdVec(2,0) = -0.180478-0.182540*I ; expctdVec(2,1) = 0.182895+0.010497*I ; expctdVec(2,2) = -0.212813-0.924801*I ;
    
    
    SECTION( "The highest level function work") { // NOT UNITARY
        HmMatrix<complex<double>> eVal(3);
        HmMatrix<complex<double>> eVec(3);
        diagonalize(A,eVec,eVal);
        CHECK( matrixApprox(eVal, expctdVal) );
        CHECK( matrixApprox(eVec, expctdVec) );
    }
    
    SECTION( "The diagonalization is internally coherent" ){
        HmMatrix<complex<double>> eVal(3);
        HmMatrix<complex<double>> eVec(3);
        diagonalize(A,eVec,eVal);
        
        CHECK(matrixApprox( A, eVec * eVal * eVec.adjoint() ));
        CHECK(matrixApprox( eVal, eVec.adjoint() * A * eVec ));
        
        CHECK(cplxApprox(( eVec * eVec.adjoint())(0,0) , 1.0 ));
        CHECK(cplxApprox(( eVec * eVec.adjoint())(1,1) , 1.0 ));
        CHECK(cplxApprox(( eVec * eVec.adjoint())(2,2) , 1.0 ));
    }
    
    //NOT IMPLEMENTED
    //SECTION( "Throwing exceptions for non-hermitian matrices ?" ) {
    
    //}
}












